#include <boost/random.hpp>
#include <boost/random/mersenne_twister.hpp>


#include <typeinfo>
#include <deque>
#include <algorithm>
#include <set>
#include <map>
#include <queue>
#include <sys/time.h>
#include <sys/resource.h>


#include <time.h>
#include "utility.h"
#include "MISPThread.h"
#include "ISPThread.h"

#include "PtDebug.h"

using namespace boost;

double GetPeakMem(void)
{
  int who = RUSAGE_SELF;
  struct rusage usage;
  int ret;

  ret = getrusage(who, &usage);

  if(0 == ret)
    return usage.ru_maxrss / 1024.0;
  else
    return -1;
}


void ReadSnapGraph(vector<list<int> > & g, ifstream & is)
{
  set<pair<int, int> > edges;
  set<int> nodes;

  int node1, node2;

  while(is >> node1 >> node2)
    {
      edges.insert(make_pair(node1, node2));
      nodes.insert(node1);
      nodes.insert(node2);
    }

  map<int, int> nodeToIndex;
  int i = 0;
  for(set<int>::iterator itor = nodes.begin();
      itor != nodes.end();
      ++itor)
    {
      nodeToIndex[*itor] = i++;
    }

  g.resize(nodes.size());

  for(set<pair<int, int> >::iterator itor = edges.begin();
      itor != edges.end();
      ++itor)
    {
      int index1 = nodeToIndex[itor->first];
      int index2 = nodeToIndex[itor->second];

      g[index1].push_back(index2);
    }

}

void ConvertToWCRawGraph(const vector<list<int> > & g, vector<list<RawLink> > & rawGraph, vector<double> & latProb)
{
  if(latProb.empty())
    {
      Pt_Info << "empty probabilities!" << endl;
      return;
    }

  vector<int> indegree;
  indegree.resize(g.size(), 0);
  for(unsigned int i = 0; i < g.size(); ++i)
    {
      for(list<int>::const_iterator itor = g[i].begin();
          itor != g[i].end();
          ++itor)
          {
              ++indegree[*itor];
          }
    }

  rawGraph.resize(g.size());
  RawLink link;

  for(unsigned int i = 0; i < g.size(); ++i)
    {
      random_shuffle(latProb.begin(), latProb.end());
      double lambda = latProb[0];
      for(list<int>::const_iterator itor = g[i].begin();
          itor != g[i].end();
          ++itor)
        {
          link.node = *itor;
          link.ap = 1.0 / indegree[*itor];
          link.lambda = lambda;
          rawGraph[i].push_back(link);
        }
    }

}

double MultiMonteCarlo(int R, int T, const set<int> & seed, const vector<list<RawLink> > & graph)
{
  if(R <= 0)
    return 0;

  double influ = 0;
  for(int i = 0; i < R; ++i)
    {
      influ += MonteCarlo(T, seed, graph);
    }

  return influ / R;
}


int MonteCarlo(int T, const set<int> & seed, const vector<list<RawLink> > & graph)
{
  static mt19937 rng(static_cast<unsigned>(time(NULL)));

  int influ = 0;

  // first: node, second: activating time
  map<int, int> nodeToActTime;
  vector<set<int> > actTimeToNode;

  actTimeToNode.resize(T + 1);
  actTimeToNode[0] = seed;// seed is activated at time 0



  for(set<int>::const_iterator itor = seed.begin();
      itor != seed.end();
      ++itor)
    {
      nodeToActTime[*itor] = 0;
    }

  for(int i = 0; i <= T; ++i)
    {
      set<int> & s = actTimeToNode[i];
      influ += s.size(); // all nodes activated at time i are counted as influence

      // let newly activated nodes to activate other nodes
      for(set<int>::iterator itor = s.begin();
	  itor != s.end();
	  ++itor)
	{
	  // for each newly activated node *itor
	  for(list<RawLink>::const_iterator it = graph[*itor].begin();
	      it != graph[*itor].end();
	      ++it)
	    {
	      map<int, int>::iterator actItor = nodeToActTime.find(it->node);
	      if(actItor != nodeToActTime.end() &&
		 actItor->second <= i)
		continue;// this node has already been activated, ignore it...

	      bernoulli_distribution<double> bern_dist(it->ap);


	      variate_generator<mt19937 &, bernoulli_distribution<double> > bern_sampler(rng, bern_dist);

	      //	      Pt_Info << bern_sampler() << " " << typeid(bern_sampler()).name() << endl;

	      if(bern_sampler())// activate success
		{
		  poisson_distribution<int> poi_dist(it->lambda);
		  variate_generator<mt19937 &, poisson_distribution<int> > poi_sampler(rng, poi_dist);

		  //		  Pt_Info << geo_sampler() << " " << typeid(geo_sampler()).name() << endl;
		  int latency = poi_sampler() + 1;// + 0.5;
		  //		  Pt_Info << "latency " << latency << endl;

		  if(i + latency > T)
		    continue;

		  if(actItor == nodeToActTime.end())// this node is in inactivate state
		    {
		      nodeToActTime[it->node] = i + latency;
		      actTimeToNode[i + latency].insert(it->node);
		    }
		  else if(i + latency < actItor->second)// this node is in latent activate state and the new activation time is more recent
		    {
		      actTimeToNode[actItor->second].erase(it->node);// remove this node from later activating set
		      actTimeToNode[i + latency].insert(it->node);// add this node to earlier activating set
		      nodeToActTime[it->node] = i + latency;
		    }
		}
	    }
	}
    }

  return influ;
}



void FindPath(int T, int node, int usedLength, double usedWeight, set<int> & usedNodes, const vector<list<RawLink> > & graph, double theta, vector<Path> & paths)
{
  if(usedLength >= T || node < 0 || node >= graph.size() || usedWeight < theta)
     return;

  // for each edge *itor of node
  for(list<RawLink>::const_iterator itor = graph[node].begin();
      itor != graph[node].end();
      ++itor)
    {
      if(usedNodes.find(itor->node) != usedNodes.end())// this node has already been used
	continue;

      // for each possible activating time of edge *itor
      // poisson distribution for delay
      double weight = usedWeight * itor->ap * exp(- itor->lambda);//itor->p;
      for(int i = 1; i <= T; ++i)
	{

	  if(usedLength + i > T)
	    break;

	  if(weight >= theta && (usedLength + i) <= T)
	    {
	      Path path;
	      path.endNode = itor->node;
	      path.weight = weight;
	      path.length = i + usedLength;

	      paths.push_back(path);

	      usedNodes.insert(itor->node);
	      FindPath(T, itor->node, i + usedLength, weight, usedNodes, graph, theta, paths);
	      usedNodes.erase(itor->node);


	      weight *= 1.0 * itor->lambda / i;//(1 - itor->p);
	    }
	  else
	    {
	      //	      break;
	    }
	}
    }
}


// paths: endNode, time t, vecotr[influence, usedNodes]
void FindPath(int T, int node, int usedLength, double usedWeight, set<int> & usedNodes, const vector<list<RawLink> > & graph, double theta, map<int, map<int, list<pair<double, set<int> > > > > & paths)
{
  if(usedLength >= T || node < 0 || node >= graph.size() || usedWeight < theta)
    return;

  // for each edge *itor of node
  for(list<RawLink>::const_iterator itor = graph[node].begin();
      itor != graph[node].end();
      ++itor)
    {
      if(usedNodes.find(itor->node) != usedNodes.end())// this node has already been used
        continue;

      // for each possible activating time of edge *itor
      double weight = usedWeight * itor->ap * exp(- itor->lambda);
      for(int i = 1; i <= T; ++i)
        {
	  if(usedLength + i > T)
	    break;
          if(weight >= theta && (usedLength + i) <= T)
            {
              usedNodes.insert(itor->node);

              paths[itor->node][i + usedLength].push_back(make_pair(weight, usedNodes));

              FindPath(T, itor->node, i + usedLength, weight, usedNodes, graph, theta, paths);
              usedNodes.erase(itor->node);


              weight *= 1.0 * itor->lambda / i;
            }
          else
            {
            }
        }
    }
}


// paths: endNode, time t, vecotr[influence, usedNodes]
void FindPath(int T, int seed, const vector<list<RawLink> > & graph, double theta, map<int, map<int, list<pair<double, set<int> > > > > & paths)
{
  if(T < 1 || theta <= 0)
    return;


  set<int> usedNodes;
  usedNodes.insert(seed);

  FindPath(T, seed, 0, 1, usedNodes, graph, theta, paths);
}


void FindPath(int T, const set<int> & seed, const vector<list<RawLink> > & graph, double theta, vector<Path> & paths)
{
  if(T < 1 || theta <= 0)
    return;


  set<int> usedNodes = seed;

  // for each node *itor in seed
  for(set<int>::const_iterator itor = seed.begin();
      itor != seed.end();
      ++itor)
    {
      FindPath(T, *itor, 0, 1, usedNodes, graph, theta, paths);
    }
}

// influence, usedNodes
double SumProb(const list<pair<double, set<int> > > & paths)
{
  double influ = 0;
  for(list<pair<double, set<int> > >::const_iterator i = paths.begin();
      i != paths.end();
      ++i)
    {
      influ = 1 - (1 - i->first) * (1 - influ);
    }
  return influ;
}

double CalInflu(const map<int, map<int, list<pair<double, set<int> > > > > & paths)
{
  // influence of all nodes
  double influ = 0;
  for(map<int, map<int, list<pair<double, set<int> > > > >::const_iterator itor = paths.begin();
      itor != paths.end();
      ++itor)
    {
      // influence of itor->first
      double ap = 0;
      for(map<int, list<pair<double, set<int> > > >::const_iterator it = itor->second.begin();
	  it != itor->second.end();
	  ++it)
	{
	  ap += (1 - ap) * SumProb(it->second);
	  
	}
      influ += ap;
    }

  return influ;
}


double CalInflu(const vector<Path> & paths)
{
  map<int, map<int, double> > nodeToTimeToProb;
  for(unsigned int i = 0; i < paths.size(); ++i)
    {
      nodeToTimeToProb[paths[i].endNode][paths[i].length] = 1 - (1 - nodeToTimeToProb[paths[i].endNode][paths[i].length]) * (1 - paths[i].weight);
    }

  double influ = 0;
  for(map<int, map<int, double> >::iterator itor = nodeToTimeToProb.begin();
      itor != nodeToTimeToProb.end();
      ++itor)
    {
      double ap = 0;
      for(map<int, double>::iterator it = itor->second.begin();
	  it != itor->second.end();
	  ++it)
	{
	  ap += (1 - ap) * it->second;
	}
      influ += ap;
    }

  return influ;
}


double IPInflu(int T, const set<int> & seed, const vector<list<RawLink> > & graph, double theta)
{
  vector<Path> paths;
  FindPath(T, seed, graph, theta, paths);
  return CalInflu(paths);
}




// return the percentage of links not pointing to neighbors
double NeighborRatio(const set<int> & neighbors, const list<RawLink> & links)
{
  if(links.empty())
    return 1;

  int n = 0;
  for(list<RawLink>::const_iterator itor = links.begin();
      itor != links.end();
      ++itor)
    {
      if(neighbors.find(itor->node) == neighbors.end())
	++n;
    }

  return double(n) / links.size();
}



// return the percentage of links not pointing to neighbors
double UncoveredNeighborRatio(map<int, double> & neighbors, const list<RawLink> & links)
{
  if(links.empty())
    return 1;

  double n = 0;
  for(list<RawLink>::const_iterator itor = links.begin();
      itor != links.end();
      ++itor)
    {
      if(neighbors.find(itor->node) == neighbors.end())
	++n;
      else
	n += (1 - neighbors[itor->node]);
    }

  return n / links.size();
}


// remove path containing node from paths
void RemovePath(int node, map<int, map<int, list<pair<double, set<int> > > > > & allPaths)
{
  allPaths.erase(node);// remove paths ending with "node"
  for(map<int, map<int, list<pair<double, set<int> > > > >::iterator itor = allPaths.begin();
      itor != allPaths.end();
      ++itor)
    {
      for(map<int, list<pair<double, set<int> > > >::iterator it = itor->second.begin();
          it != itor->second.end();
          ++it)
        {
          for(list<pair<double, set<int> > >::iterator i = it->second.begin();
              i != it->second.end();)
            {
              if(i->second.find(node) != i->second.end())// this path contains node, thus remove it
                i = it->second.erase(i);
              else
                ++i;
            }
        }
    }
}


// remove path containing node from allPaths
// remove path containing kNodes from paths
// combine them into allPaths
void UnionPaths(map<int, map<int, list<pair<double, set<int> > > > > & allPaths, 
		set<int> & kNodes,
		const map<int, map<int, list<pair<double, set<int> > > > > & paths,
		int node)
{
  // remove path containing node from allPaths
  RemovePath(node, allPaths);
  // remove path containing kNodes from paths
  map<int, map<int, list<pair<double, set<int> > > > > tempPaths = paths;
  for(set<int>::iterator itor = kNodes.begin();
      itor != kNodes.end();
      ++itor)
    {
      RemovePath(*itor, tempPaths);
    }

  for(map<int, map<int, list<pair<double, set<int> > > > >::iterator itor = tempPaths.begin();
      itor != tempPaths.end();
      ++itor)
    {
      for(map<int, list<pair<double, set<int> > > >::iterator it = itor->second.begin();
	  it != itor->second.end();
	  ++it)
	{
	  list<pair<double, set<int> > > & ls = allPaths[itor->first][it->first];

      ls.insert(ls.begin(), it->second.begin(), it->second.end());

	}
    }
}


// paths: endNode, time t, vecotr[influence, usedNodes]
void SetCombIPCelfKNodes(int T, const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes)
{
  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  vector<map<int, map<int, list<pair<double, set<int> > > > > > nodeToPaths;
  nodeToPaths.resize(graph.size());


  //void FindPath(int T, int seed, const vector<list<RawLink> > & graph, double theta, map<int, map<int, vector<pair<double, set<int> > > > > & paths)
  for(int i = 0; i < graph.size(); ++i)
    {
      FindPath(T, i, graph, theta, nodeToPaths[i]);
      double influ = CalInflu(nodeToPaths[i]);
      q.push(make_pair(influ, make_pair(i, 0)));
    }


  double originalInflu = 0;

  // all paths starting from current seed set
  map<int, map<int, list<pair<double, set<int> > > > > allPaths;


  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
	      originalInflu += element.first;
	      UnionPaths(allPaths, kNodes, nodeToPaths[element.second.first], element.second.first);
              kNodes.insert(element.second.first);

              break;
            }
	  else// this delta influence is invalid
            {
              element.second.second = i;
	      map<int, map<int, list<pair<double, set<int> > > > > tempPaths = allPaths;
	      UnionPaths(tempPaths, kNodes, nodeToPaths[element.second.first], element.second.first);
	      element.first = CalInflu(tempPaths) - originalInflu;// set valid delta influence

              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }

}


void RatioWeightedCombIPCelfKNodes(int T, const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes)
{
  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  vector<double> nodeToInflu;
  nodeToInflu.resize(graph.size());

  for(int i = 0; i < graph.size(); ++i)
    {
      set<int> s;
      s.insert(i);
      double influ = IPInflu(T, s, graph, theta);
      nodeToInflu[i] = influ;
      q.push(make_pair(influ, make_pair(i, 0)));
    }


  map<int, double> nodeToCoveredRatio;

  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
              kNodes.insert(element.second.first);

	      // update the covered ratio of add all neighbors of the newly added seed node
	      for(list<RawLink>::const_iterator itor = graph[element.second.first].begin();
		  itor != graph[element.second.first].end();
		  ++itor)
		{
		  map<int, double>::iterator it = nodeToCoveredRatio.find(itor->node);
		  if(it == nodeToCoveredRatio.end())
		    nodeToCoveredRatio[itor->node] = itor->ap;
		  else
		    {
		      it->second = 1 - (1 - it->second) * (1 - itor->ap);
		    }
		}

              break;
            }
          else// this delta influence is invalid
            {
              element.second.second = i;
              element.first = nodeToInflu[element.second.first] * UncoveredNeighborRatio(nodeToCoveredRatio, graph[element.second.first]);
              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }
}



void RatioWeightedCombIPCelfKNodesMT(int T, const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes, int threadNum)
{
  if(threadNum < 1)
    return;

  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  vector<double> nodeToInflu;
  nodeToInflu.resize(graph.size());


  if(1 == threadNum)
    {
      for(int i = 0; i < graph.size(); ++i)
	{
	  set<int> s;
	  s.insert(i);
	  double influ = IPInflu(T, s, graph, theta);
	  nodeToInflu[i] = influ;
	  q.push(make_pair(influ, make_pair(i, 0)));
	}
    }
  else
    {
      int idx = 0;
      boost::mutex idxLock;
      boost::mutex qLock;
      MISPThread * workers[threadNum];
      for(int i = 0; i < threadNum; ++i)
	{
	  workers[i] = new MISPThread(idx, idxLock, q, qLock, T, graph, theta, nodeToInflu);
	  if(!workers[i]->Start())
	    {
	      Pt_Info << "start worker error ... " << i << endl;
	    }
	}

      for(int i = 0; i < threadNum; ++i)
	{
	  while(!workers[i]->IsThreadExited())
	    {
	      boost::this_thread::sleep(boost::posix_time::seconds(2));
	    }
	  delete workers[i];
	}
    }


  map<int, double> nodeToCoveredRatio;

  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
              kNodes.insert(element.second.first);

              // update the covered ratio of add all neighbors of the newly added seed node
              for(list<RawLink>::const_iterator itor = graph[element.second.first].begin();
                  itor != graph[element.second.first].end();
                  ++itor)
                {
                  map<int, double>::iterator it = nodeToCoveredRatio.find(itor->node);
                  if(it == nodeToCoveredRatio.end())
                    nodeToCoveredRatio[itor->node] = itor->ap;
                  else
                    {
                      it->second = 1 - (1 - it->second) * (1 - itor->ap);
                    }
                }

              break;
            }
	  else// this delta influence is invalid
            {
              element.second.second = i;
              element.first = nodeToInflu[element.second.first] * UncoveredNeighborRatio(nodeToCoveredRatio, graph[element.second.first]);
              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }
}


void IPCelfKNodes(int T, const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes)
{
  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;


  for(int i = 0; i < graph.size(); ++i)
    {
      set<int> s;
      s.insert(i);
      double influ = IPInflu(T, s, graph, theta);
      q.push(make_pair(influ, make_pair(i, 0)));
    }



  double originalInflu = 0;

  for(int i = 0; i < K; ++i)
    {
      while(true)
	{
	  pair<double, pair<int, int> > element = q.top();
	  q.pop();

	  if(element.second.second == i)// this is a valid delta influence
	    {
	      kNodes.insert(element.second.first);
	      originalInflu += element.first;
	      break;
	    }
	  else// this delta influence is invalid 
	    {
	      element.second.second = i;
	      set<int> s = kNodes;
	      s.insert(element.second.first);
	      element.first = IPInflu(T, s, graph, theta) - originalInflu;
	      q.push(element);
	    }
	}
      Pt_Info << i << " nodes selected!" << endl;
    }

}


void IPCelfKNodesMT(int T, const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes, int threadNum)
{

  if(threadNum < 1)
    return;

  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  if(1 == threadNum)
    {
      for(int i = 0; i < graph.size(); ++i)
        {
          set<int> s;
          s.insert(i);
          double influ = IPInflu(T, s, graph, theta);
          q.push(make_pair(influ, make_pair(i, 0)));
        }
    }
  else
    {
      int idx = 0;
      boost::mutex idxLock;
      boost::mutex qLock;
      ISPThread * workers[threadNum];
      for(int i = 0; i < threadNum; ++i)
        {
          workers[i] = new ISPThread(idx, idxLock, q, qLock, T, graph, theta);
          if(!workers[i]->Start())
            {
              Pt_Info << "start worker error ... " << i << endl;
            }
        }

      for(int i = 0; i < threadNum; ++i)
        {
          while(!workers[i]->IsThreadExited())
            {
	      boost::this_thread::sleep(boost::posix_time::seconds(2));
            }
          delete workers[i];
        }
    }



  double originalInflu = 0;

  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
              kNodes.insert(element.second.first);
              originalInflu += element.first;
              break;
            }
          else// this delta influence is invalid
            {
              element.second.second = i;
              set<int> s = kNodes;
              s.insert(element.second.first);
              element.first = IPInflu(T, s, graph, theta) - originalInflu;
              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }

}



void MCCelfKNodes(int T, const vector<list<RawLink> > & graph, int K, set<int> & kNodes, int R)
{
  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;



  for(int i = 0; i < graph.size(); ++i)
    {
      set<int> s;
      s.insert(i);
      double influ = MultiMonteCarlo(R, T, s, graph);
      q.push(make_pair(influ, make_pair(i, 0)));
    }



  double originalInflu = 0;

  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
              kNodes.insert(element.second.first);
              originalInflu += element.first;
              break;
            }
          else// this delta influence is invalid
            {
              element.second.second = i;
              set<int> s = kNodes;
              s.insert(element.second.first);
              element.first = MultiMonteCarlo(R, T, s, graph) - originalInflu;
              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }

}


void RandomKNodes(const vector<list<RawLink> > & graph, int K, set<int> & kNodes)
{
  vector<int> v;
  for(unsigned int i = 0; i < graph.size(); ++i)
    v.push_back(i);

  random_shuffle(v.begin(), v.end());

  for(int i = 0; i < K; ++i)
    kNodes.insert(v[i]);
}

void LargeDegreeKNodes(const vector<list<RawLink> > & graph, int K, set<int> & kNodes)
{
  if(K < 1)
    return;

  // degree, node id
  set<pair<int, int> > s;
  for(int i = 0; i < graph.size(); ++i)
    {
      int degree = graph[i].size();

      if(s.size() < K)
	{
	  s.insert(make_pair(degree, i));
	}
      else
	{
	  if(degree > s.begin()->first)
	    {
	      s.erase(s.begin());
	      s.insert(make_pair(degree, i));
	    }
	}
    }

  for(set<pair<int, int> >::iterator itor = s.begin();
      itor != s.end();
      ++itor)
    {
      kNodes.insert(itor->second);
    }
}


void LargeIPKNodes(int T, const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes)
{
  if(K < 1)
    return;

  // influence, node id
  set<pair<double, int> > s;
  for(int i = 0; i < graph.size(); ++i)
    {
      set<int> seed;
      seed.insert(i);
      double influ = IPInflu(T, seed, graph, theta);

      if(s.size() < K)
        {
          s.insert(make_pair(influ, i));
        }
      else
        {
          if(influ > s.begin()->first)
            {
              s.erase(s.begin());
              s.insert(make_pair(influ, i));
            }
        }
    }

  for(set<pair<double, int> >::iterator itor = s.begin();
      itor != s.end();
      ++itor)
    {
      kNodes.insert(itor->second);
    }
}



