#include <boost/random.hpp>
#include <boost/random/mersenne_twister.hpp>


#include <typeinfo>
#include <deque>
#include <algorithm>
#include <set>
#include <map>
#include <queue>
#include <sys/time.h>
#include <sys/resource.h>


#include <time.h>
#include "utility.h"
#include "MISPThread.h"
#include "ISPThread.h"

#include "PtDebug.h"

using namespace boost;

double GetPeakMem(void)
{
  int who = RUSAGE_SELF;
  struct rusage usage;
  int ret;

  ret = getrusage(who, &usage);

  if(0 == ret)
    return usage.ru_maxrss / 1024.0;
  else
    return -1;
}


void ReadSnapGraph(vector<list<int> > & g, ifstream & is)
{
  set<pair<int, int> > edges;
  set<int> nodes;

  int node1, node2;

  while(is >> node1 >> node2)
    {
      edges.insert(make_pair(node1, node2));
      nodes.insert(node1);
      nodes.insert(node2);
    }

  map<int, int> nodeToIndex;
  int i = 0;
  for(set<int>::iterator itor = nodes.begin();
      itor != nodes.end();
      ++itor)
    {
      nodeToIndex[*itor] = i++;
    }

  g.resize(nodes.size());

  for(set<pair<int, int> >::iterator itor = edges.begin();
      itor != edges.end();
      ++itor)
    {
      int index1 = nodeToIndex[itor->first];
      int index2 = nodeToIndex[itor->second];

      g[index1].push_back(index2);
    }

}

void ReadFlixsterGraph(vector<list<int> > & g, ifstream & is)
{
  set<pair<int, int> > edges;
  set<int> nodes;

  int node1, node2;

  while(is >> node1 >> node2)
    {
      if(node1 < node2)
	{
	  edges.insert(make_pair(node1, node2));
	}
      else if(node1 > node2)
	{
	  edges.insert(make_pair(node2, node1));
	}
      else
	{
	  Pt_Info << "find a self edge : " << node1 << " " << node2 << endl;
	}
      nodes.insert(node1);
      nodes.insert(node2);
    }


  map<int, int> nodeToIndex;
  int i = 0;
  for(set<int>::iterator itor = nodes.begin();
      itor != nodes.end();
      ++itor)
    {
      nodeToIndex[*itor] = i++;
    }

  g.resize(nodes.size());

  for(set<pair<int, int> >::iterator itor = edges.begin();
      itor != edges.end();
      ++itor)
    {
      int index1 = nodeToIndex[itor->first];
      int index2 = nodeToIndex[itor->second];

      g[index1].push_back(index2);
      g[index2].push_back(index1);
    }
}

void ConvertToWCRawGraph(const vector<list<int> > & g, vector<list<RawLink> > & rawGraph)
{
  vector<int> indegree;
  indegree.resize(g.size(), 0);
  for(unsigned int i = 0; i < g.size(); ++i)
    {
      for(list<int>::const_iterator itor = g[i].begin();
          itor != g[i].end();
          ++itor)
	{
	  ++indegree[*itor];
	}
    }

  rawGraph.resize(g.size());
  RawLink link;

  for(unsigned int i = 0; i < g.size(); ++i)
    {
      for(list<int>::const_iterator itor = g[i].begin();
          itor != g[i].end();
          ++itor)
        {
          link.node = *itor;
          link.ap = 1.0 / indegree[*itor];
          rawGraph[i].push_back(link);
        }
    }

}

double MultiMonteCarlo(int R, const set<int> & seed, const vector<list<RawLink> > & graph)
{
  if(R <= 0)
    return 0;

  double influ = 0;
  for(int i = 0; i < R; ++i)
    {
      influ += MonteCarlo(seed, graph);
    }

  return influ / R;
}


int MonteCarlo(const set<int> & seed, const vector<list<RawLink> > & graph)
{
  static mt19937 rng(static_cast<unsigned>(time(NULL)));

  set<int> activatedNodes(seed);
  queue<int> q;

  for(set<int>::const_iterator itor = seed.begin();
      itor != seed.end();
      ++itor)
    {
      q.push(*itor);
    }


  while(!q.empty())
    {
      int node = q.front();
      q.pop();

      for(list<RawLink>::const_iterator it = graph[node].begin();
	  it != graph[node].end();
	  ++it)
	{
	  if(activatedNodes.find(it->node) != activatedNodes.end())// this node has already be activated
	    continue;

	  bernoulli_distribution<double> bern_dist(it->ap);
	  variate_generator<mt19937 &, bernoulli_distribution<double> > bern_sampler(rng, bern_dist);

	  if(bern_sampler())// activate success
	    {
	      activatedNodes.insert(it->node);
	      q.push(it->node);
	    }
	}
    }


  return activatedNodes.size();
}



void FindPath(int node, double usedWeight, set<int> & usedNodes, const vector<list<RawLink> > & graph, double theta, vector<Path> & paths)
{
  if(node < 0 || node >= graph.size() || usedWeight < theta)
     return;

  // for each edge *itor of node
  for(list<RawLink>::const_iterator itor = graph[node].begin();
      itor != graph[node].end();
      ++itor)
    {
      if(usedNodes.find(itor->node) != usedNodes.end())// this node has already been used
	continue;

      // for each possible activating time of edge *itor
      // poisson distribution for delay
      double weight = usedWeight * itor->ap;

      if(weight >= theta)
	{
	  Path path;
	  path.endNode = itor->node;
	  path.weight = weight;

	  paths.push_back(path);
	  usedNodes.insert(itor->node);
	  FindPath(itor->node, weight, usedNodes, graph, theta, paths);
	  usedNodes.erase(itor->node);
	}
    }
}

void FindPath(const set<int> & seed, const vector<list<RawLink> > & graph, double theta, vector<Path> & paths)
{
  if(theta <= 0)
    return;


  set<int> usedNodes = seed;

  // for each node *itor in seed
  for(set<int>::const_iterator itor = seed.begin();
      itor != seed.end();
      ++itor)
    {
      FindPath(*itor, 1, usedNodes, graph, theta, paths);
    }
}

double CalInflu(const vector<Path> & paths)
{
  map<int, double> nodeToProb;
  for(unsigned int i = 0; i < paths.size(); ++i)
    {
      nodeToProb[paths[i].endNode] = 1 - (1 - nodeToProb[paths[i].endNode]) * (1 - paths[i].weight);
    }


  double influ = 0;
  for(map<int, double>::iterator itor = nodeToProb.begin();
      itor != nodeToProb.end();
      ++itor)
    {
      influ += itor->second;
    }

  return influ;
}


double IPInflu(const set<int> & seed, const vector<list<RawLink> > & graph, double theta)
{
  vector<Path> paths;
  FindPath(seed, graph, theta, paths);
  return CalInflu(paths);
}

// return the percentage of links not pointing to neighbors
double UncoveredNeighborRatio(map<int, double> & neighbors, const list<RawLink> & links)
{
  if(links.empty())
    return 1;

  double n = 0;
  for(list<RawLink>::const_iterator itor = links.begin();
      itor != links.end();
      ++itor)
    {
      if(neighbors.find(itor->node) == neighbors.end())
	++n;
      else
	n += (1 - neighbors[itor->node]);
    }

  return n / links.size();
}

void RatioWeightedCombIPCelfKNodes(const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes)
{
  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  vector<double> nodeToInflu;
  nodeToInflu.resize(graph.size());

  for(int i = 0; i < graph.size(); ++i)
    {
      set<int> s;
      s.insert(i);
      double influ = IPInflu(s, graph, theta);
      nodeToInflu[i] = influ;
      q.push(make_pair(influ, make_pair(i, 0)));
    }

  map<int, double> nodeToCoveredRatio;

  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
              kNodes.insert(element.second.first);

	      // update the covered ratio of add all neighbors of the newly added seed node
	      for(list<RawLink>::const_iterator itor = graph[element.second.first].begin();
		  itor != graph[element.second.first].end();
		  ++itor)
		{
		  map<int, double>::iterator it = nodeToCoveredRatio.find(itor->node);
		  if(it == nodeToCoveredRatio.end())
		    nodeToCoveredRatio[itor->node] = itor->ap;
		  else
		    {
		      it->second = 1 - (1 - it->second) * (1 - itor->ap);
		    }
		}

              break;
            }
          else// this delta influence is invalid
            {
              element.second.second = i;
              element.first = nodeToInflu[element.second.first] * UncoveredNeighborRatio(nodeToCoveredRatio, graph[element.second.first]);
              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }
}



void RatioWeightedCombIPCelfKNodesMT(const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes, int threadNum)
{
  if(threadNum < 1)
    return;

  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  vector<double> nodeToInflu;
  nodeToInflu.resize(graph.size());


  if(1 == threadNum)
    {
      for(int i = 0; i < graph.size(); ++i)
	{
	  set<int> s;
	  s.insert(i);
	  double influ = IPInflu(s, graph, theta);
	  nodeToInflu[i] = influ;
	  q.push(make_pair(influ, make_pair(i, 0)));
	}
    }
  else
    {
      int idx = 0;
      boost::mutex idxLock;
      boost::mutex qLock;
      MISPThread * workers[threadNum];
      for(int i = 0; i < threadNum; ++i)
	{
	  workers[i] = new MISPThread(idx, idxLock, q, qLock, graph, theta, nodeToInflu);
	  if(!workers[i]->Start())
	    {
	      Pt_Info << "start worker error ... " << i << endl;
	    }
	}

      for(int i = 0; i < threadNum; ++i)
	{
	  while(!workers[i]->IsThreadExited())
	    {
	      boost::this_thread::sleep(boost::posix_time::seconds(2));
	    }
	  delete workers[i];
	}
    }


  map<int, double> nodeToCoveredRatio;

  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
              kNodes.insert(element.second.first);

              // update the covered ratio of add all neighbors of the newly added seed node
              for(list<RawLink>::const_iterator itor = graph[element.second.first].begin();
                  itor != graph[element.second.first].end();
                  ++itor)
                {
                  map<int, double>::iterator it = nodeToCoveredRatio.find(itor->node);
                  if(it == nodeToCoveredRatio.end())
                    nodeToCoveredRatio[itor->node] = itor->ap;
                  else
                    {
                      it->second = 1 - (1 - it->second) * (1 - itor->ap);
                    }
                }

              break;
            }
	  else// this delta influence is invalid
            {
              element.second.second = i;
              element.first = nodeToInflu[element.second.first] * UncoveredNeighborRatio(nodeToCoveredRatio, graph[element.second.first]);
              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }
}


void IPCelfKNodesMT(const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes, int threadNum)
{
  if(threadNum < 1)
    return;

  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  if(1 == threadNum)
    {
      for(int i = 0; i < graph.size(); ++i)
        {
          set<int> s;
          s.insert(i);
          double influ = IPInflu(s, graph, theta);
          q.push(make_pair(influ, make_pair(i, 0)));
        }
    }
  else
    {
      int idx = 0;
      boost::mutex idxLock;
      boost::mutex qLock;
      ISPThread * workers[threadNum];
      for(int i = 0; i < threadNum; ++i)
        {
          workers[i] = new ISPThread(idx, idxLock, q, qLock, graph, theta);
          if(!workers[i]->Start())
            {
              Pt_Info << "start worker error ... " << i << endl;
            }
        }

      for(int i = 0; i < threadNum; ++i)
        {
          while(!workers[i]->IsThreadExited())
            {
	      boost::this_thread::sleep(boost::posix_time::seconds(2));
            }
          delete workers[i];
        }
    }


  double originalInflu = 0;

  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
              kNodes.insert(element.second.first);
              originalInflu += element.first;
              break;
            }
          else// this delta influence is invalid
            {
              element.second.second = i;
              set<int> s = kNodes;
              s.insert(element.second.first);
              element.first = IPInflu(s, graph, theta) - originalInflu;
              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }


}

void IPCelfKNodes(const vector<list<RawLink> > & graph, double theta, int K, set<int> & kNodes)
{
  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  for(int i = 0; i < graph.size(); ++i)
    {
      set<int> s;
      s.insert(i);
      double influ = IPInflu(s, graph, theta);
      q.push(make_pair(influ, make_pair(i, 0)));
    }



  double originalInflu = 0;

  for(int i = 0; i < K; ++i)
    {
      while(true)
	{
	  pair<double, pair<int, int> > element = q.top();
	  q.pop();

	  if(element.second.second == i)// this is a valid delta influence
	    {
	      kNodes.insert(element.second.first);
	      originalInflu += element.first;
	      break;
	    }
	  else// this delta influence is invalid 
	    {
	      element.second.second = i;
	      set<int> s = kNodes;
	      s.insert(element.second.first);
	      element.first = IPInflu(s, graph, theta) - originalInflu;
	      q.push(element);
	    }
	}
      Pt_Info << i << " nodes selected!" << endl;
    }

}




void MCCelfKNodes(const vector<list<RawLink> > & graph, int K, set<int> & kNodes, int R)
{
  // delta influence, node id, size of original seed set
  priority_queue<pair<double, pair<int, int> > > q;

  for(int i = 0; i < graph.size(); ++i)
    {
      set<int> s;
      s.insert(i);
      double influ = MultiMonteCarlo(R, s, graph);
      q.push(make_pair(influ, make_pair(i, 0)));
    }

  double originalInflu = 0;

  for(int i = 0; i < K; ++i)
    {
      while(true)
        {
          pair<double, pair<int, int> > element = q.top();
          q.pop();

          if(element.second.second == i)// this is a valid delta influence
            {
              kNodes.insert(element.second.first);
              originalInflu += element.first;
              break;
            }
          else// this delta influence is invalid
            {
              element.second.second = i;
              set<int> s = kNodes;
              s.insert(element.second.first);
              element.first = MultiMonteCarlo(R, s, graph) - originalInflu;
              q.push(element);
            }
        }
      Pt_Info << i << " nodes selected!" << endl;
    }

}


void RandomKNodes(const vector<list<RawLink> > & graph, int K, set<int> & kNodes)
{
  vector<int> v;
  for(unsigned int i = 0; i < graph.size(); ++i)
    v.push_back(i);

  random_shuffle(v.begin(), v.end());

  for(int i = 0; i < K; ++i)
    kNodes.insert(v[i]);
}
